# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail model importer extensions, for CORE-POS integration
"""

import decimal

from rattail import importing
from rattail.util import pretty_quantity


##############################
# core importer overrides
##############################

class PersonImporter(importing.model.PersonImporter):

    extensions = {
        '_corepos': [
            'corepos_customer_id',
        ],
    }

    def cache_query(self):
        query = super(PersonImporter, self).cache_query()
        model = self.config.get_model()

        # we want to ignore people with no CORE ID, if that's (part of) our key
        if 'corepos_customer_id' in self.key:
            query = query.join(model.CorePerson)\
                         .filter(model.CorePerson.corepos_customer_id != None)

        return query


class CustomerImporter(importing.model.CustomerImporter):

    extensions = {
        '_corepos': [
            'corepos_account_id',
            'corepos_card_number',
        ],
    }


class CustomerShopperImporter(importing.model.CustomerShopperImporter):

    extensions = {
        '_corepos': [
            'corepos_customer_id',
        ],
    }


class MemberImporter(importing.model.MemberImporter):

    extensions = {
        '_corepos': [
            'corepos_account_id',
        ],
    }


class MemberEquityPaymentImporter(importing.model.MemberEquityPaymentImporter):

    extensions = {
        '_corepos': [
            'corepos_card_number',
            'corepos_transaction_number',
            'corepos_transaction_id',
            'corepos_department_number',
            'corepos_datetime',
        ],
    }


class StoreImporter(importing.model.StoreImporter):

    extensions = {
        '_corepos': [
            'corepos_id',
        ],
    }


class DepartmentImporter(importing.model.DepartmentImporter):

    extensions = {
        '_corepos': [
            'corepos_number',
        ],
    }


class SubdepartmentImporter(importing.model.SubdepartmentImporter):

    extensions = {
        '_corepos': [
            'corepos_number',
        ],
    }


class VendorImporter(importing.model.VendorImporter):

    extensions = {
        '_corepos': [
            'corepos_id',
        ],
    }


class ProductImporter(importing.model.ProductImporter):

    extensions = {
        '_corepos': [
            'corepos_id',
        ],
    }

    def setup(self):
        super(ProductImporter, self).setup()

        if self.fields_active(self.size_fields):
            app = self.config.get_app()
            handler = app.get_products_handler()
            self.uoms = handler.get_uom_sil_codes(self.session, uppercase=True)

    def cache_query(self):
        query = super(ProductImporter, self).cache_query()
        model = self.config.get_model()

        # we want to ignore products with no CORE ID, if that's (part of) our key
        if 'corepos_id' in self.key:
            query = query.join(model.CoreProduct)\
                         .filter(model.CoreProduct.corepos_id != None)

        return query

    def get_uom_code(self, uom):
        if hasattr(self, 'uoms'):
            return self.uoms.get(uom.upper())

        app = self.config.get_app()
        handler = app.get_products_handler()
        return handler.get_uom_sil_code(self.session, uom.upper())

    def normalize_size_info(self, core_product):

        # convert product to dict if needed
        if isinstance(core_product, dict):
            core_data = core_product
        else:
            core_data = {
                'size': core_product.size,
                'unitofmeasure': core_product.unit_of_measure,
            }

        unit_size = None
        if 'size' in core_data and core_data['size'] is not None:
            try:
                unit_size = decimal.Decimal(core_data['size'])
            except decimal.InvalidOperation:
                pass

        uom_abbrev = core_data.get('unitofmeasure')

        uom_code = self.enum.UNIT_OF_MEASURE_NONE
        if uom_abbrev is not None:
            uom_code = self.get_uom_code(uom_abbrev) or self.enum.UNIT_OF_MEASURE_NONE

        if unit_size is not None and uom_abbrev is not None:
            size = "{} {}".format(pretty_quantity(unit_size), uom_abbrev)
        elif unit_size is not None:
            size = pretty_quantity(unit_size)
        elif uom_abbrev is not None:
            size = uom_abbrev
        else:
            size = None

        return {
            'size': size,
            'unit_size': unit_size,
            'uom_abbrev': uom_abbrev,
            'uom_code': uom_code,
        }


class ProductCostImporter(importing.model.ProductCostImporter):

    extensions = {
        '_corepos': [
            'corepos_id',
        ],
    }

    def cache_query(self):
        query = super(ProductCostImporter, self).cache_query()
        model = self.config.get_model()

        # we want to ignore items with no CORE ID, if that's (part of) our key
        if 'corepos_id' in self.key:
            query = query.join(model.CoreProductCost)\
                         .filter(model.CoreProductCost.corepos_id != None)

        return query



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS commands
"""

import sys
import warnings

from rattail import commands
from rattail_corepos import __version__
from rattail_corepos.corepos.office.commands import ExportCSV, ImportCSV


def main(*args):
    """
    Primary entry point for Crepes command system
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)


class Command(commands.Command):
    """
    Primary command for Crepes (CORE-POS)
    """
    name = 'crepes'
    version = __version__
    description = "Crepes -- command line interface for CORE-POS"
    long_description = ""


class ImportToCore(commands.ImportSubcommand):
    """
    Generic base class for commands which import *to* a CORE DB.
    """


class ExportCore(commands.ImportSubcommand):
    """
    Export data to another CORE database
    """
    name = 'export-core'
    description = __doc__.strip()
    handler_key = 'to_corepos_db_office_op.from_corepos_db_office_op.export'
    default_dbkey = 'host'

    def add_parser_args(self, parser):
        super(ExportCore, self).add_parser_args(parser)
        parser.add_argument('--dbkey', metavar='KEY', default=self.default_dbkey,
                            help="Config key for database engine to be used as the \"target\" "
                            "CORE DB, i.e. where data will be exported.  This key must be "
                            "defined in the [rattail_corepos.db] section of your config file.")

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs


class LegacyExportCSV(ExportCSV):

    def __init__(self, *args, **kwargs):
        warnings.warn("the `crepes export-csv` command is deprecated; "
                      "please use `core-office export-csv` instead",
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)


class ImportCore(ImportToCore):
    """
    Import data from another CORE database
    """
    name = 'import-core'
    description = __doc__.strip()
    handler_key = 'to_corepos_db_office_op.from_corepos_db_office_op.import'
    accepts_dbkey_param = True

    def add_parser_args(self, parser):
        super(ImportCore, self).add_parser_args(parser)
        if self.accepts_dbkey_param:
            parser.add_argument('--dbkey', metavar='KEY', default='host',
                                help="Config key for database engine to be used as the CORE "
                                "\"host\", i.e. the source of the data to be imported.  This key "
                                "must be defined in the [rattail_corepos.db] section of your config file.  "
                                "Defaults to 'host'.")

    def get_handler_kwargs(self, **kwargs):
        if self.accepts_dbkey_param:
            if 'args' in kwargs:
                kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs


class LegacyImportCSV(ImportCSV):

    def __init__(self, *args, **kwargs):
        warnings.warn("the `crepes import-csv` command is deprecated; "
                      "please use `core-office import-csv` instead",
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Schema for CORE member update batch
"""

import sqlalchemy as sa

from rattail.db import model
from rattail.db.core import filename_column


class CoreMemberBatch(model.BatchMixin, model.Base):
    """
    Hopefully generic batch for adding / updating member data in CORE.
    """
    batch_key = 'corepos_member'
    __tablename__ = 'batch_corepos_member'
    __batchrow_class__ = 'CoreMemberBatchRow'
    model_title = "CORE Member Batch"
    model_title_plural = "CORE Member Batches"

    STATUS_OK                           = 1
    STATUS_CANNOT_PARSE_FILE            = 2

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_CANNOT_PARSE_FILE        : "cannot parse file",
    }

    input_file = filename_column(nullable=True, doc="""
    Base name of the input data file.
    """)


class CoreMemberBatchRow(model.BatchRowMixin, model.Base):
    """
    Row of data within a CORE member batch.
    """
    __tablename__ = 'batch_corepos_member_row'
    __batch_class__ = CoreMemberBatch

    STATUS_NO_CHANGE                    = 1
    STATUS_MEMBER_NOT_FOUND             = 2
    STATUS_FIELDS_CHANGED               = 3

    STATUS = {
        STATUS_NO_CHANGE                : "no change",
        STATUS_MEMBER_NOT_FOUND         : "member not found",
        STATUS_FIELDS_CHANGED           : "update member",
    }

    card_number_raw = sa.Column(sa.String(length=20), nullable=True)
    card_number = sa.Column(sa.Integer(), nullable=True)

    first_name = sa.Column(sa.String(length=30), nullable=True)
    first_name_old = sa.Column(sa.String(length=30), nullable=True)
    last_name = sa.Column(sa.String(length=30), nullable=True)
    last_name_old = sa.Column(sa.String(length=30), nullable=True)

    street = sa.Column(sa.String(length=255), nullable=True)
    street_old = sa.Column(sa.String(length=255), nullable=True)
    city = sa.Column(sa.String(length=20), nullable=True)
    city_old = sa.Column(sa.String(length=20), nullable=True)
    state = sa.Column(sa.String(length=2), nullable=True)
    state_old = sa.Column(sa.String(length=2), nullable=True)
    zipcode = sa.Column(sa.String(length=10), nullable=True)
    zipcode_old = sa.Column(sa.String(length=10), nullable=True)

    phone = sa.Column(sa.String(length=30), nullable=True)
    phone_old = sa.Column(sa.String(length=30), nullable=True)

    email1 = sa.Column(sa.String(length=50), nullable=True)
    email1_old = sa.Column(sa.String(length=50), nullable=True)
    email2 = sa.Column(sa.String(length=50), nullable=True)
    email2_old = sa.Column(sa.String(length=50), nullable=True)

    member_type_id = sa.Column(sa.SmallInteger(), nullable=True)
    member_type_id_old = sa.Column(sa.SmallInteger(), nullable=True)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS model importers (direct DB)

.. warning::
   All classes in this module are "direct DB" importers, which will write
   directly to MySQL.  They are meant to be used in dry-run mode only, and/or
   for sample data import to a dev system etc.  They are *NOT* meant for
   production use, as they will completely bypass any CORE business rules logic
   which may exist.
"""

import logging

from corepos.db.office_op import model as corepos, Session as CoreSession
from corepos.db.office_trans import model as coretrans, Session as CoreTransSession

from rattail import importing
from rattail.importing.handlers import ToSQLAlchemyHandler


log = logging.getLogger(__name__)


class ToCoreHandler(ToSQLAlchemyHandler):
    """
    Base class for import handlers which target a CORE database on the local side.
    """
    # TODO: should change local_key in all apps! since dbtype can vary now
    local_key = 'corepos_db_office_op'
    generic_local_title = "CORE Office (DB)"
    corepos_dbtype = 'office_op'
    corepos_dbkey = 'default'

    @property
    def local_title(self):
        dbtype = 'op'
        if self.corepos_dbtype == 'office_trans':
            dbtype = 'trans'
        return f"CORE Office (DB '{dbtype}')"

    def make_session(self):

        # session type depends on the --corepos-dbtype arg
        if self.corepos_dbtype == 'office_trans':
            return CoreTransSession(bind=self.config.coretrans_engines[self.corepos_dbkey])

        # assume office_op by default
        return CoreSession(bind=self.config.corepos_engines[self.corepos_dbkey])


class ToCore(importing.ToSQLAlchemy):
    """
    Base class for all CORE "operational" model importers.
    """

    def create_object(self, key, host_data):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return host_data

        return super(ToCore, self).create_object(key, host_data)

    def update_object(self, obj, host_data, **kwargs):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return obj

        return super(ToCore, self).update_object(obj, host_data, **kwargs)

    def delete_object(self, obj):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return True

        return super(ToCore, self).delete_object(obj)


class ToCoreTrans(importing.ToSQLAlchemy):
    """
    Base class for all CORE "transaction" model importers
    """


########################################
# CORE Operational
########################################

class DepartmentImporter(ToCore):
    model_class = corepos.Department
    key = 'number'


class SubdepartmentImporter(ToCore):
    model_class = corepos.Subdepartment
    key = 'number'


class VendorImporter(ToCore):
    model_class = corepos.Vendor
    key = 'id'


class VendorContactImporter(ToCore):
    model_class = corepos.VendorContact
    key = 'vendor_id'


class ProductImporter(ToCore):
    model_class = corepos.Product
    key = 'id'


class ProductFlagImporter(ToCore):
    model_class = corepos.ProductFlag
    key = 'bit_number'


class VendorItemImporter(ToCore):
    model_class = corepos.VendorItem
    key = ('sku', 'vendor_id')


class EmployeeImporter(ToCore):
    model_class = corepos.Employee
    key = 'number'


class CustomerClassicImporter(ToCore):
    model_class = corepos.CustomerClassic
    key = 'id'


class MemberTypeImporter(ToCore):
    model_class = corepos.MemberType
    key = 'id'


class MemberInfoImporter(ToCore):
    model_class = corepos.MemberInfo
    key = 'card_number'

    # default is to read first/last name directly from `meminfo`
    # table, but if this flag is set then names will be read from
    # `custdata` instead (where `personNum = 1`) although `meminfo`
    # will still be used as fallback in that case.
    prefer_local_names_from_custdata = False

    # default is to compare street address "as-is" between systems.
    # but CORE can be inconsistent with presence or lack of newline
    # suffix, when address is just one line (which is the typical
    # scenario).  so if this flag is set then importer will strip
    # whitespace automatically when reading this field value.
    strip_address_street = False

    # similar to previous flag, this one makes the importer strip
    # whitespace from *all* address fields when reading.
    strip_address_all = False

    # somewhat akin to the address flags, this one if set will cause
    # the importer to "normalize" phone numbers read from CORE.  the
    # assumption would be that the host/source side of the import
    # would also be presenting normalized phone data for comparison to
    # work correctly.  cf. AppHandler.normalize_phone_number()
    normalize_phone_numbers = False

    @property
    def supported_fields(self):
        fields = list(super(MemberInfoImporter, self).supported_fields)

        fields.append('member_type_id')

        return fields

    def normalize_local_object(self, member):
        data = super(MemberInfoImporter, self).normalize_local_object(member)

        if self.prefer_local_names_from_custdata and member.customers:
            customer = member.customers[0]
            data['first_name'] = customer.first_name
            data['last_name'] = customer.last_name

        if self.normalize_phone_numbers and 'phone' in self.fields:
            data['phone'] = self.app.normalize_phone_number(data['phone'])

        if ((self.strip_address_street or self.strip_address_all)
            and 'street' in self.fields):
            data['street'] = (data['street'] or '').strip()
        if self.strip_address_all and self.fields_active(['city', 'state', 'zip']):
            data['city'] = (data['city'] or '').strip()
            data['state'] = (data['state'] or '').strip()
            data['zip'] = (data['zip'] or '').strip()

        if 'member_type_id' in self.fields:
            data['member_type_id'] = None
            customer = member.customers[0] if member.customers else None
            if customer:
                data['member_type_id'] = customer.member_type_id

        return data

    def update_object(self, member, host_data, local_data=None, **kwargs):
        member = super(MemberInfoImporter, self).update_object(
            member, host_data, local_data=local_data, **kwargs)

        if 'first_name' in self.fields:
            if member.customers:
                customer = member.customers[0]
                first_name = host_data['first_name']
                if customer.first_name != first_name:
                    customer.first_name = first_name

        if 'last_name' in self.fields:
            if member.customers:
                customer = member.customers[0]
                last_name = host_data['last_name']
                if customer.last_name != last_name:
                    customer.last_name = last_name

        if 'member_type_id' in self.fields:
            member_type_id = host_data['member_type_id']
            for customer in member.customers:
                if customer.member_type_id != member_type_id:
                    if member_type_id is None:
                        log.warning("will not blank out member_type_id for "
                                    "member #%s: %s %s",
                                    member.card_number,
                                    customer.first_name,
                                    customer.last_name)
                    else:
                        customer.member_type_id = member_type_id

        return member


class MemberDateImporter(ToCore):
    model_class = corepos.MemberDate
    key = 'card_number'


class MemberContactImporter(ToCore):
    model_class = corepos.MemberContact
    key = 'card_number'


class HouseCouponImporter(ToCore):
    model_class = corepos.HouseCoupon
    key = 'coupon_id'


########################################
# CORE Transactions
########################################

class StockPurchaseImporter(ToCoreTrans):
    """
    CORE-POS stock purchase data importer.
    """
    model_class = coretrans.StockPurchase


class TransactionDetailImporter(ToCoreTrans):
    """
    CORE-POS transaction data importer.
    """
    model_class = coretrans.TransactionDetail



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail Commands
"""

from rattail import commands
from rattail.util import load_object


class ExportCore(commands.ImportSubcommand):
    """
    Export data from Rattail to CORE-POS
    """
    name = 'export-corepos'
    description = __doc__.strip()
    handler_key = 'to_corepos_api.from_rattail.export'


class ImportCOREPOSAPI(commands.ImportSubcommand):
    """
    Import data from a CORE POS API
    """
    name = 'import-corepos-api'
    description = __doc__.strip()
    handler_key = 'to_rattail.from_corepos_api.import'


class ImportCOREPOSDB(commands.ImportSubcommand):
    """
    Import data from a CORE POS database
    """
    name = 'import-corepos-db'
    description = __doc__.strip()
    handler_key = 'to_rattail.from_corepos_db_office_op.import'

    def add_parser_args(self, parser):
        super().add_parser_args(parser)

        parser.add_argument('--corepos-dbtype', metavar='TYPE', default='office_op',
                            choices=['office_op', 'office_trans'],
                            help="Config *type* for CORE-POS database engine to be used as "
                            "host.  Default type is 'office_op' - this determines which "
                            "config section is used with regard to the --corepos-dbkey arg.")

        parser.add_argument('--corepos-dbkey', metavar='KEY', default='default',
                            help="Config key for CORE POS database engine to be used as the \"host\", "
                            "i.e. the source of the data to be imported.  This key " "must be "
                            "defined in the [rattail_corepos.db] section of your config file.  "
                            "Defaults to 'default'.")

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['corepos_dbtype'] = kwargs['args'].corepos_dbtype
            kwargs['corepos_dbkey'] = kwargs['args'].corepos_dbkey
        return kwargs


class CoreImportSquare(commands.ImportFromCSV):
    """
    Import transaction data from Square into CORE
    """
    name = 'corepos-import-square'
    description = __doc__.strip()
    handler_key = 'to_corepos_db_office_trans.from_square_csv.import'



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync for CORE POS
"""

import sqlalchemy as sa

from corepos.db.office_op import Session as CoreSession, model as corepos

from rattail.db import model
from rattail.datasync import DataSyncWatcher, DataSyncImportConsumer


class CoreOfficeOpWatcher(DataSyncWatcher):
    """
    DataSync watcher for the CORE ``office_op`` database.
    """
    prunes_changes = True

    def __init__(self, *args, **kwargs):
        super(CoreOfficeOpWatcher, self).__init__(*args, **kwargs)

        self.changes_table_name = kwargs.get('changes_table_name',
                                             'datasync_changes')

        self.corepos_metadata = sa.MetaData()
        self.corepos_changes = sa.Table(
            self.changes_table_name, self.corepos_metadata,
            sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
            sa.Column('object_type', sa.String(length=255), nullable=False),
            sa.Column('object_key', sa.String(length=255), nullable=False),
            sa.Column('deleted', sa.Boolean(), nullable=False, default=False))

    def get_changes(self, lastrun):
        session = CoreSession()
        result = session.execute(self.corepos_changes.select())
        changes = result.fetchall()
        session.close()
        if changes:
            return [
                (c.id,
                 model.DataSyncChange(
                     payload_type=c.object_type,
                     payload_key=c.object_key,
                     deletion=c.deleted))
                for c in changes]

    def prune_changes(self, keys):
        deleted = 0
        session = CoreSession()
        for key in keys:
            result = session.execute(self.corepos_changes.select()\
                                     .where(self.corepos_changes.c.id == key))
            if result.fetchall():
                session.execute(self.corepos_changes.delete()\
                                .where(self.corepos_changes.c.id == key))
                deleted += 1
        session.commit()
        session.close()
        return deleted


class COREPOSProductWatcher(DataSyncWatcher):
    """
    DataSync watcher for the CORE POS database.
    """

    def get_changes(self, lastrun):
        if not lastrun:
            return

        changes = []
        session = CoreSession()
        lastrun = self.localize_lastrun(session, lastrun)

        # Department
        departments = session.query(corepos.Department)\
                             .filter(corepos.Department.modified >= lastrun)\
                             .all()
        if departments:
            changes.extend([
                (None,
                 model.DataSyncChange(
                     payload_type='Department',
                     payload_key=str(dept.number)))
                for dept in departments])

        # TODO: subdepartment table doesn't have a modified flag?
        # # Subdepartment
        # subdepartments = session.query(corepos.Subdepartment)\
        #                         .filter(corepos.Subdepartment.modified >= lastrun)\
        #                         .all()
        # if subdepartments:
        #     changes.extend([
        #         (None,
        #          model.DataSyncChange(
        #              payload_type='Subdepartment',
        #              payload_key=six.text_type(subdept.subdept_no)))
        #         for subdept in subdepartments])

        # TODO: vendor table doesn't have a modified flag?
        # # Vendor
        # vendors = session.query(corepos.Vendor)\
        #                  .filter(corepos.Vendor.modified >= lastrun)\
        #                  .all()
        # if vendors:
        #     changes.extend([
        #         (None,
        #          model.DataSyncChange(
        #              payload_type='Vendor',
        #              payload_key=six.text_type(vendor.vendorID)))
        #         for vendor in vendors])

        # Product
        products = session.query(corepos.Product)\
                          .filter(corepos.Product.modified >= lastrun)\
                          .all()
        if products:
            changes.extend([
                (None,
                 model.DataSyncChange(
                     payload_type='Product',
                     payload_key=product.upc))
                for product in products
                if product.upc])

        session.close()
        return changes


class FromRattailToCore(DataSyncImportConsumer):
    """
    Rattail -> CORE POS datasync consumer
    """
    handler_spec = 'rattail_corepos.corepos.importing.rattail:FromRattailToCore'

    def process_changes(self, session, changes):
        """
        Process all the given changes, coming from Rattail.
        """
        # TODO: this probably doesn't accomplish anything here?
        if self.runas_username:
            session.set_continuum_user(self.runas_username)

        # update all importers with current session
        for importer in self.importers.values():
            importer.host_session = session
            # also establish the API client for each!
            importer.establish_api()

        # sync all Member changes
        types = [
            'Customer',
            'CustomerShopper',
            'Person',
            'CustomerPerson',
            'CustomerMailingAddress',
            'PersonPhoneNumber',
            'PersonEmailAddress',
            'Member',
        ]
        for change in [c for c in changes if c.payload_type in types]:
            if change.payload_type == 'Customer' and change.deletion:
                # # just do default logic for this one
                # self.invoke_importer(session, change)
                # TODO: we have no way to delete a CORE customer via API, right?
                pass
            else: # we consider this an "add/update"
                customers = self.get_customers(session, change)
                for customer in customers:
                    self.process_change(session, self.importers['Member'],
                                        host_object=customer)

        # sync all Vendor changes
        types = [
            'Vendor',
            'VendorPhoneNumber',
            'VendorEmailAddress',
        ]
        for change in [c for c in changes if c.payload_type in types]:
            if change.payload_type == 'Vendor' and change.deletion:
                # # just do default logic for this one
                # self.invoke_importer(session, change)
                # TODO: we have no way to delete a CORE vendor via API, right?
                pass
            else: # we consider this an "add/update"
                vendor = self.get_vendor(session, change)
                if vendor:
                    self.process_change(session, self.importers['Vendor'],
                                        host_object=vendor)

        # sync all Product changes
        types = [
            'Product',
            'ProductPrice',
        ]
        for change in [c for c in changes if c.payload_type in types]:
            if change.payload_type == 'Product' and change.deletion:
                # # just do default logic for this one
                # self.invoke_importer(session, change)
                # TODO: we have no way to delete a CORE product via API, right?
                pass
            else: # we consider this an "add/update"
                product = self.get_product(session, change)
                if product:
                    self.process_change(session, self.importers['Product'],
                                        host_object=product)

        # process all remaining supported models with typical logic
        types = [
            'Department',
            'Subdepartment',
        ]
        for change in [c for c in changes if c.payload_type in types]:
            self.invoke_importer(session, change)

    def get_host_object(self, session, change):
        return session.get(getattr(model, change.payload_type), change.payload_key)

    def get_customers(self, session, change):
        clientele = self.app.get_clientele_handler()

        if change.payload_type == 'Customer':
            customer = session.get(model.Customer, change.payload_key)
            if customer:
                return [customer]

        elif change.payload_type == 'CustomerShopper':
            shopper = session.get(model.CustomerShopper, change.payload_key)
            if shopper:
                return [shopper.customer]

        elif change.payload_type == 'CustomerPerson':
            cp = session.get(model.CustomerPerson, change.payload_key)
            if cp:
                return [cp.customer]

        elif change.payload_type == 'Person':
            person = session.get(model.Person, change.payload_key)
            if person:
                return clientele.get_customers_for_account_holder(person)

        elif change.payload_type == 'CustomerMailingAddress':
            address = session.get(model.CustomerMailingAddress, change.payload_key)
            if address:
                return [address.customer]

        elif change.payload_type == 'PersonPhoneNumber':
            phone = session.get(model.PersonPhoneNumber, change.payload_key)
            if phone:
                return clientele.get_customers_for_account_holder(phone.person)

        elif change.payload_type == 'PersonEmailAddress':
            email = session.get(model.PersonEmailAddress, change.payload_key)
            if email:
                return clientele.get_customers_for_account_holder(email.person)

        elif change.payload_type == 'Member':
            member = session.get(model.Member, change.payload_key)
            if member:
                return [member.customer]

        return []

    def get_vendor(self, session, change):

        if change.payload_type == 'Vendor':
            return session.get(model.Vendor, change.payload_key)

        if change.payload_type == 'VendorPhoneNumber':
            phone = session.get(model.VendorPhoneNumber, change.payload_key)
            if phone:
                return phone.vendor

        if change.payload_type == 'VendorEmailAddress':
            email = session.get(model.VendorEmailAddress, change.payload_key)
            if email:
                return email.vendor

    def get_product(self, session, change):

        if change.payload_type == 'Product':
            return session.get(model.Product, change.payload_key)

        if change.payload_type == 'ProductPrice':
            price = session.get(model.ProductPrice, change.payload_key)
            if price:
                return price.product



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Square -> CORE-POS data importing
"""

import re
import datetime
import decimal
from collections import OrderedDict

import sqlalchemy as sa

from corepos.db.office_trans import Session as CoreTransSession, model as coretrans

from rattail import importing
from rattail_corepos.corepos.office.importing import db as corepos_importing


class FromSquareToCoreTrans(importing.ToSQLAlchemyHandler):
    """
    Square -> CORE-POS import handler.
    """
    host_title = "Square"
    local_title = "CORE-POS"

    def make_session(self):
        return CoreTransSession()

    def get_importers(self):
        importers = OrderedDict()
        importers['TransactionDetail'] = TransactionDetailImporter
        return importers


class FromSquare(importing.FromCSV):
    """
    Base class for Square -> CORE-POS importers.
    """


class TransactionDetailImporter(FromSquare, corepos_importing.model.TransactionDetailImporter):
    """
    Transaction detail importer.
    """
    key = 'store_row_id'
    supported_fields = [
        'store_row_id',
        'date_time',
        'card_number',
        'upc',
        'description',
        'quantity',
        'unit_price',
        'discount',
        'tax',
        'total',
    ]

    batches_supported = True

    def setup(self):
        super(TransactionDetailImporter, self).setup()

        # cache existing transactions by ID
        self.transaction_details = self.cache_model(coretrans.TransactionDetail,
                                                    key=self.transaction_detail_key)

        # keep track of new IDs
        self.new_ids = {}
        self.last_new_id = self.get_last_new_id()

    def transaction_detail_key(self, detail, normal):
        return (
            detail.store_id,
            detail.register_number,
            detail.date_time,
            detail.upc,
        )

    def get_last_new_id(self):
        # TODO: pretty sure there is a better way to do this...
        return self.session.query(sa.func.max(coretrans.TransactionDetail.store_row_id))\
                           .scalar() or 0

    currency_pattern = re.compile(r'^\$(?P<amount>\d+\.\d\d)$')
    currency_pattern_negative = re.compile(r'^\(\$(?P<amount>\d+\.\d\d)\)$')

    def parse_currency(self, value):
        value = (value or '').strip() or None
        if value:

            # first check for positive amount
            match = self.currency_pattern.match(value)
            if match:
                return float(match.group('amount'))

            # okay then, check for negative amount
            match = self.currency_pattern_negative.match(value)
            if match:
                return 0 - float(match.group('amount'))

    def normalize_host_object(self, csvrow):

        # date_time
        date = datetime.datetime.strptime(csvrow['Date'], '%m/%d/%Y').date()
        time = datetime.datetime.strptime(csvrow['Time'], '%H:%M:%S').time()
        date_time = datetime.datetime.combine(date, time)

        # upc
        upc = csvrow['SKU']

        # store_row_id
        key = (
            0,                  # store_id
            None,               # register_number
            date_time,
            upc,
        )
        if key in self.transaction_details:
            store_row_id = self.transaction_details[key].store_row_id
        else:
            store_row_id = self.last_new_id + 1
            self.new_ids[store_row_id] = csvrow
            self.last_new_id = store_row_id

        # card_number
        card_number = csvrow['Customer Reference ID'] or None
        if card_number:
            card_number = int(card_number)

        # description
        description = csvrow['Item']

        # quantity
        quantity = float(csvrow['Qty'])

        # unit_price
        unit_price = self.parse_currency(csvrow['Gross Sales'])
        if unit_price is not None:
            unit_price /= quantity
            unit_price = decimal.Decimal('{:0.2f}'.format(unit_price))
        elif csvrow['Gross Sales']:
            log.warning("cannot parse 'unit_price' from: %s", csvrow['Gross Sales'])

        # discount
        discount = self.parse_currency(csvrow['Discounts'])
        if discount is not None:
            discount = decimal.Decimal('{:0.2f}'.format(discount))
        elif csvrow['Discounts']:
            log.warning("cannot parse 'discount' from: %s", csvrow['Discounts'])

        # tax
        tax = self.parse_currency(csvrow['Tax'])
        if csvrow['Tax'] and tax is None:
            log.warning("cannot parse 'tax' from: %s", csvrow['Tax'])
        tax = bool(tax)

        # total
        total = self.parse_currency(csvrow['Net Sales'])
        if total is not None:
            total = decimal.Decimal('{:0.2f}'.format(total))
        elif csvrow['Net Sales']:
            log.warning("cannot parse 'total' from: %s", csvrow['Net Sales'])

        return {
            '_object_str': "({}) {}".format(upc, description),
            'store_row_id': store_row_id,
            'date_time': date_time,
            'card_number': card_number,
            'upc': upc,
            'description': description,
            'quantity': quantity,
            'unit_price': unit_price,
            'discount': discount,
            'tax': tax,
            'total': total,
        }



# -*- coding: utf-8; -*-
"""add equity_payment.corepos_datetime

Revision ID: d426a274f0c2
Revises: 93978a7adc66
Create Date: 2023-09-13 20:59:42.706994

"""

# revision identifiers, used by Alembic.
revision = 'd426a274f0c2'
down_revision = '93978a7adc66'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_member_equity_payment
    op.add_column('corepos_member_equity_payment', sa.Column('corepos_datetime', sa.DateTime(), nullable=True))
    op.add_column('corepos_member_equity_payment_version', sa.Column('corepos_datetime', sa.DateTime(), autoincrement=False, nullable=True))


def downgrade():

    # corepos_member_equity_payment
    op.drop_column('corepos_member_equity_payment_version', 'corepos_datetime')
    op.drop_column('corepos_member_equity_payment', 'corepos_datetime')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Schema for CORE equity import batch
"""

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db import model


class CoreEquityImportBatch(model.BatchMixin, model.Base):
    """
    Hopefully generic batch for CORE equity import
    """
    batch_key = 'corepos_equity_import'
    __tablename__ = 'batch_corepos_equity_import'
    __batchrow_class__ = 'CoreEquityImportBatchRow'
    model_title = "CORE-POS Equity Import Batch"
    model_title_plural = "CORE-POS Equity Import Batches"

    STATUS_OK                           = 1

    STATUS = {
        STATUS_OK                       : "ok",
    }


class CoreEquityImportBatchRow(model.BatchRowMixin, model.Base):
    """
    Row of data within a CORE equity import batch.
    """
    __tablename__ = 'batch_corepos_equity_import_row'
    __batch_class__ = CoreEquityImportBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['payment_uuid'], ['member_equity_payment.uuid'],
                                    name='batch_corepos_equity_import_row_fk_payment'),
            sa.ForeignKeyConstraint(['member_uuid'], ['member.uuid'],
                                    name='batch_corepos_equity_import_row_fk_member'),
        )

    STATUS_OK                           = 1
    STATUS_MEMBER_NOT_FOUND             = 2
    STATUS_MISSING_VALUES               = 3
    STATUS_NEEDS_ATTENTION              = 4
    STATUS_ALREADY_IN_CORE              = 5
    STATUS_EQUITY_OVERPAID              = 6

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_MEMBER_NOT_FOUND         : "member not found",
        STATUS_MISSING_VALUES           : "missing values",
        STATUS_NEEDS_ATTENTION          : "needs attention",
        STATUS_ALREADY_IN_CORE          : "already in CORE-POS",
        STATUS_EQUITY_OVERPAID          : "equity overpaid",
    }

    payment_uuid = sa.Column(sa.String(length=32), nullable=True)
    payment = orm.relationship(model.MemberEquityPayment)

    member_uuid = sa.Column(sa.String(length=32), nullable=True)
    member = orm.relationship(model.Member)

    card_number = sa.Column(sa.Integer(), nullable=True)
    first_name = sa.Column(sa.String(length=30), nullable=True)
    last_name = sa.Column(sa.String(length=30), nullable=True)
    member_type_id = sa.Column(sa.Integer(), nullable=True)
    payment_amount = sa.Column(sa.Numeric(precision=10, scale=2), nullable=True)
    department_number = sa.Column(sa.Integer(), nullable=True)
    tender_code = sa.Column(sa.String(length=2), nullable=True)
    timestamp = sa.Column(sa.DateTime(), nullable=True)
    corepos_equity_total = sa.Column(sa.Numeric(precision=10, scale=2), nullable=True)
    rattail_equity_total = sa.Column(sa.Numeric(precision=10, scale=2), nullable=True)
    other_equity_total = sa.Column(sa.Numeric(precision=10, scale=2), nullable=True)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Handler supplement
"""

from rattail.app import RattailProvider, GenericHandler


class CoreProvider(RattailProvider):
    """
    App provider for CORE-POS integration.
    """

    def get_corepos_handler(self, **kwargs):
        if not hasattr(self, 'corepos_handler'):
            spec = self.config.get('rattail', 'corepos.handler',
                                   default='rattail_corepos.app:CoreHandler')
            factory = self.load_object(spec)
            self.corepos_handler = factory(self.config, **kwargs)
        return self.corepos_handler


class CoreHandler(GenericHandler):
    """
    Handler for CORE-POS integration.
    """

    def get_office_url(
            self,
            require=False,
            **kwargs):
        """
        Returns the base URL for the CORE Office web app.
        """
        getter = self.config.require if require else self.config.get
        url = getter('corepos', 'office.url')
        if url:
            return url.rstrip('/')

    def get_office_member_url(
            self,
            card_number,
            office_url=None,
            require=False,
            **kwargs):
        """
        Returns the CORE Office URL for the customer account with the
        given card number.
        """
        if not office_url:
            office_url = self.get_office_url(require=require)
        if office_url:
            return f'{office_url}/mem/MemberEditor.php?memNum={card_number}'

    # TODO: deprecate / remove this
    get_office_customer_account_url = get_office_member_url

    def get_office_product_url(
            self,
            upc,
            office_url=None,
            require=False,
            **kwargs):
        """
        Returns the CORE Office URL for the product with the given
        UPC.
        """
        if not office_url:
            office_url = self.get_office_url(require=require)
        if office_url:
            return f'{office_url}/item/ItemEditorPage.php?searchupc={upc}'

    def get_office_vendor_url(
            self,
            id,
            office_url=None,
            require=False,
            **kwargs):
        """
        Returns the CORE Office URL for the vendor with the given ID.
        """
        if not office_url:
            office_url = self.get_office_url(require=require)
        if office_url:
            return f'{office_url}/item/vendors/VendorIndexPage.php?vid={id}'



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS -> CORE-POS data import
"""

from collections import OrderedDict

from corepos.db.office_op import Session as CoreSession

from rattail.importing.handlers import FromSQLAlchemyHandler
from rattail.importing.sqlalchemy import FromSQLAlchemySameToSame
from rattail_corepos.corepos.office.importing import db as corepos_importing


class FromCoreHandler(FromSQLAlchemyHandler):
    """
    Base class for import handlers which use a CORE database as the host / source.
    """
    host_title = "CORE"
    host_key = 'corepos_db_office_op'

    def make_host_session(self):
        return CoreSession()


# TODO: deprecate / remove this
ToCoreHandler = corepos_importing.model.ToCoreHandler


class FromCoreToCoreBase(object):
    """
    Common base class for Core -> Core data import/export handlers.
    """

    def get_importers(self):
        importers = OrderedDict()
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Vendor'] = VendorImporter
        importers['VendorContact'] = VendorContactImporter
        importers['Product'] = ProductImporter
        importers['ProductFlag'] = ProductFlagImporter
        importers['VendorItem'] = VendorItemImporter
        importers['Employee'] = EmployeeImporter
        importers['CustomerClassic'] = CustomerClassicImporter
        importers['MemberType'] = MemberTypeImporter
        importers['MemberInfo'] = MemberInfoImporter
        importers['HouseCoupon'] = HouseCouponImporter
        return importers


class FromCoreToCoreImport(FromCoreToCoreBase, FromCoreHandler, corepos_importing.model.ToCoreHandler):
    """
    Handler for CORE (other) -> CORE (local) data import.

    .. attribute:: direction

       Value is ``'import'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    dbkey = 'host'
    local_title = "CORE (default)"

    @property
    def host_title(self):
        return "CORE ({})".format(self.dbkey)

    def make_host_session(self):
        return CoreSession(bind=self.config.corepos_engines[self.dbkey])


class FromCoreToCoreExport(FromCoreToCoreBase, FromCoreHandler, corepos_importing.model.ToCoreHandler):
    """
    Handler for CORE (local) -> CORE (other) data export.

    .. attribute:: direction

       Value is ``'export'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    direction = 'export'
    host_title = "CORE (default)"

    @property
    def local_title(self):
        return "CORE ({})".format(self.dbkey)

    def make_session(self):
        return CoreSession(bind=self.config.corepos_engines[self.dbkey])


class FromCore(FromSQLAlchemySameToSame):
    """
    Base class for CORE -> CORE data importers.
    """


class DepartmentImporter(FromCore, corepos_importing.model.DepartmentImporter):
    pass

class SubdepartmentImporter(FromCore, corepos_importing.model.SubdepartmentImporter):
    pass

class VendorImporter(FromCore, corepos_importing.model.VendorImporter):
    pass

class VendorContactImporter(FromCore, corepos_importing.model.VendorContactImporter):
    pass

class ProductImporter(FromCore, corepos_importing.model.ProductImporter):
    pass

class ProductFlagImporter(FromCore, corepos_importing.model.ProductFlagImporter):
    pass

class VendorItemImporter(FromCore, corepos_importing.model.VendorItemImporter):
    pass

class EmployeeImporter(FromCore, corepos_importing.model.EmployeeImporter):
    pass

class CustomerClassicImporter(FromCore, corepos_importing.model.CustomerClassicImporter):
    pass

class MemberTypeImporter(FromCore, corepos_importing.model.MemberTypeImporter):
    pass

class MemberInfoImporter(FromCore, corepos_importing.model.MemberInfoImporter):
    pass

class HouseCouponImporter(FromCore, corepos_importing.model.HouseCouponImporter):
    pass



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office utility logic
"""

import json
import os
import subprocess

from rattail.files import resource_path


def get_fannie_config_value(config, name):
    """
    Retrieve a config value from `fannie/config.php`

    :param config: Rattail config object.

    :param name: Name of the config value to be returned.  This need
       not be prefixed with ``FANNIE_`` although it can be.

    :returns: Config value as Python object, e.g. a string or dict.
       This is believed to work okay but since the Fannie config file
       represents data with PHP code, which is then converted to JSON
       and finally to Python, some discrepancies may be possible.
    """
    if not name.startswith('FANNIE_'):
        name = f'FANNIE_{name}'

    is4c = config.require('corepos', 'srcdir')
    path = os.path.join(is4c, 'fannie', 'config.php')
    script = resource_path('rattail_corepos.corepos.office:scripts/parse-fannie-config.php')

    try:
        output = subprocess.check_output(['php', script,
                                          '--path', path,
                                          '--setting', name])
    except subprocess.CalledProcessError as error:
        raise ValueError(f"failed to read value: {error.output.decode('utf_8')}")

    return json.loads(output.decode('utf_8'))



# -*- coding: utf-8; -*-
"""more model extensions

Revision ID: 9c5029effe93
Revises: c61f78243ff3
Create Date: 2020-07-06 19:27:06.156401

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '9c5029effe93'
down_revision = 'c61f78243ff3'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_customer
    op.create_table('corepos_customer',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_account_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['customer.uuid'], name='corepos_customer_fk_customer'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_customer_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_account_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_customer_version_end_transaction_id'), 'corepos_customer_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_customer_version_operation_type'), 'corepos_customer_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_customer_version_transaction_id'), 'corepos_customer_version', ['transaction_id'], unique=False)

    # corepos_member
    op.create_table('corepos_member',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_account_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['member.uuid'], name='corepos_member_fk_member'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_member_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_account_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_member_version_end_transaction_id'), 'corepos_member_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_member_version_operation_type'), 'corepos_member_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_member_version_transaction_id'), 'corepos_member_version', ['transaction_id'], unique=False)

    # corepos_department
    op.create_table('corepos_department',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_number', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['department.uuid'], name='corepos_department_fk_department'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_department_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_department_version_end_transaction_id'), 'corepos_department_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_department_version_operation_type'), 'corepos_department_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_department_version_transaction_id'), 'corepos_department_version', ['transaction_id'], unique=False)

    # corepos_subdepartment
    op.create_table('corepos_subdepartment',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_number', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['subdepartment.uuid'], name='corepos_subdepartment_fk_subdepartment'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_subdepartment_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_subdepartment_version_end_transaction_id'), 'corepos_subdepartment_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_subdepartment_version_operation_type'), 'corepos_subdepartment_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_subdepartment_version_transaction_id'), 'corepos_subdepartment_version', ['transaction_id'], unique=False)

    # corepos_product
    op.create_table('corepos_product',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['product.uuid'], name='corepos_product_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_product_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_product_version_end_transaction_id'), 'corepos_product_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_product_version_operation_type'), 'corepos_product_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_product_version_transaction_id'), 'corepos_product_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_product
    op.drop_index(op.f('ix_corepos_product_version_transaction_id'), table_name='corepos_product_version')
    op.drop_index(op.f('ix_corepos_product_version_operation_type'), table_name='corepos_product_version')
    op.drop_index(op.f('ix_corepos_product_version_end_transaction_id'), table_name='corepos_product_version')
    op.drop_table('corepos_product_version')
    op.drop_table('corepos_product')

    # corepos_subdepartment
    op.drop_index(op.f('ix_corepos_subdepartment_version_transaction_id'), table_name='corepos_subdepartment_version')
    op.drop_index(op.f('ix_corepos_subdepartment_version_operation_type'), table_name='corepos_subdepartment_version')
    op.drop_index(op.f('ix_corepos_subdepartment_version_end_transaction_id'), table_name='corepos_subdepartment_version')
    op.drop_table('corepos_subdepartment_version')
    op.drop_table('corepos_subdepartment')

    # corepos_department
    op.drop_index(op.f('ix_corepos_department_version_transaction_id'), table_name='corepos_department_version')
    op.drop_index(op.f('ix_corepos_department_version_operation_type'), table_name='corepos_department_version')
    op.drop_index(op.f('ix_corepos_department_version_end_transaction_id'), table_name='corepos_department_version')
    op.drop_table('corepos_department_version')
    op.drop_table('corepos_department')

    # corepos_member
    op.drop_index(op.f('ix_corepos_member_version_transaction_id'), table_name='corepos_member_version')
    op.drop_index(op.f('ix_corepos_member_version_operation_type'), table_name='corepos_member_version')
    op.drop_index(op.f('ix_corepos_member_version_end_transaction_id'), table_name='corepos_member_version')
    op.drop_table('corepos_member_version')
    op.drop_table('corepos_member')

    # corepos_customer
    op.drop_index(op.f('ix_corepos_customer_version_transaction_id'), table_name='corepos_customer_version')
    op.drop_index(op.f('ix_corepos_customer_version_operation_type'), table_name='corepos_customer_version')
    op.drop_index(op.f('ix_corepos_customer_version_end_transaction_id'), table_name='corepos_customer_version')
    op.drop_table('corepos_customer_version')
    op.drop_table('corepos_customer')



# -*- coding: utf-8; -*-
"""add card_number_raw for member import

Revision ID: d6a0f21a6a94
Revises: 50961b4b854a
Create Date: 2022-03-15 11:24:41.764317

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'd6a0f21a6a94'
down_revision = '50961b4b854a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_corepos_member_row
    op.add_column('batch_corepos_member_row', sa.Column('card_number_raw', sa.String(length=20), nullable=True))


def downgrade():

    # batch_corepos_member_row
    op.drop_column('batch_corepos_member_row', 'card_number_raw')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS model importers (webservices API)
"""

from rattail import importing
from rattail.util import data_diffs
from rattail_corepos.corepos.util import get_core_members
from rattail_corepos.corepos.api import make_corepos_api


class ToCOREAPIHandler(importing.ImportHandler):
    """
    Base class for handlers targeting the CORE API.
    """
    local_key = 'corepos_api'
    generic_local_title = "CORE Office (API)"

    @property
    def local_title(self):
        return "CORE-POS (API)"

    def process_changes(self, changes):

        if 'Member' in changes:
            member_changes = changes['Member']
            member_importer = self.importers['Member']
            created, updated, deleted = changes['Member']
            if updated:

                # explode the 'customers' field so email will show
                # diffs for each subrecord field
                for update in updated:
                    local_object, local_data, host_data = update

                    local_customers = local_data.pop('customers')
                    for i, customer in enumerate(local_customers, 1):
                        for key in customer:
                            local_data[f'customers.{i}.{key}'] = customer[key]

                    host_customers = host_data.pop('customers')
                    for i, customer in enumerate(host_customers, 1):
                        for key in customer:
                            host_data[f'customers.{i}.{key}'] = customer[key]

        # do normal logic for the rest
        super().process_changes(changes)


class ToCoreAPI(importing.Importer):
    """
    Base class for all CORE "operational" model importers, which use the API.
    """
    # TODO: these importers are in a bit of an experimental state at the
    # moment.  we only allow create/update b/c it will use the API instead of
    # direct DB
    allow_delete = False

    caches_local_data = True

    def setup(self):
        self.establish_api()

    def establish_api(self):
        self.api = make_corepos_api(self.config)

    def ensure_fields(self, data):
        """
        Ensure each of our supported fields are included in the data.  This is
        to handle cases where the API does not return all fields, e.g. when
        some of them are empty.
        """
        for field in self.fields:
            if field not in data:
                data[field] = None

    def fix_empties(self, data, fields):
        """
        Fix "empty" values for the given set of fields.  This just uses an
        empty string instead of ``None`` for each, to add some consistency
        where the API might lack it.

        Main example so far, is the Vendor API, which may not return some
        fields at all (and so our value is ``None``) in some cases, but in
        other cases it *will* return a value, default of which is the empty
        string.  So we want to "pretend" that we get an empty string back even
        when we actually get ``None`` from it.
        """
        for field in fields:
            if data[field] is None:
                data[field] = ''


class MemberImporter(ToCoreAPI):
    """
    Member model importer for CORE-POS
    """
    model_name = 'Member'
    key = 'cardNo'
    supported_fields = [
        'cardNo'
        'customerAccountID',
        'customers',
        # 'memberStatus',
        # 'activeStatus',
        # 'customerTypeID',
        # 'chargeBalance',
        # 'chargeLimit',
        # 'idCardUPC',
        'startDate',
        'endDate',
        'addressFirstLine',
        'addressSecondLine',
        'city',
        'state',
        'zip',
        # 'contactAllowed',
        # 'contactMethod',
        # 'modified',
    ]
    supported_customer_fields = [
        'customerID',
        # 'customerAccountID',
        # 'cardNo',
        'firstName',
        'lastName',
        # 'chargeAllowed',
        # 'checksAllowed',
        # 'discount',
        'accountHolder',
        # 'staff',
        'phone',
        'altPhone',
        'email',
        # 'memberPricingAllowed',
        # 'memberCouponsAllowed',
        # 'lowIncomeBenefits',
        # 'modified',
    ]

    #empty_date_value = '0000-00-00 00:00:00'
    empty_date_value = '1900-01-01 00:00:00'

    def get_local_objects(self, host_data=None):
        return get_core_members(self.config, self.api, progress=self.progress)

    def get_single_local_object(self, key):
        assert len(self.key) == 1
        assert self.key[0] == 'cardNo'
        return self.api.get_member(key[0])

    def normalize_local_object(self, member):
        data = dict(member)
        return data

    def data_diffs(self, local_data, host_data):
        diffs = super(MemberImporter, self).data_diffs(local_data, host_data)

        # the 'customers' field requires a more granular approach, since the
        # data coming from API may have different fields than our local data
        if 'customers' in self.fields and 'customers' in diffs:
            if not self.customer_data_differs(local_data, host_data):
                diffs.remove('customers')

        # # also the start/end dates should be looked at more closely.  if they
        # # contain the special '__omit__' value then we won't ever count as diff
        # if 'startDate' in self.fields and 'startDate' in diffs:
        #     if host_data['startDate'] == '__omit__':
        #         diffs.remove('startDate')
        # if 'endDate' in self.fields and 'endDate' in diffs:
        #     if host_data['endDate'] == '__omit__':
        #         diffs.remove('endDate')

        return diffs

    def customer_data_differs(self, local_data, host_data):
        local_customers = local_data['customers']
        host_customers = host_data['customers']

        # if both are empty, we're good
        if not local_customers and not host_customers:
            return False

        # obviously we differ if record count doesn't match
        if len(local_customers) != len(host_customers):
            return True

        # okay then, let's traverse the "new" list
        for host_customer in host_customers:

            # we differ if can't locate corresponding "old" local record
            local_customer = self.find_local_customer(local_customers, host_customer)
            if not local_customer:
                return True

            # we differ if old and new records differ, but...
            diffs = data_diffs(local_customer, host_customer,
                               fields=self.supported_customer_fields)

            # nb. the customerID field is sometimes int, sometimes str
            # (?)  so we must coerce both sides to str for proper
            # comparison, otherwise false diffs happen
            if diffs and (
                    diffs != ['customerID']
                    or str(local_customer['customerID']) != str(host_customer['customerID'])):
                return True

        # okay, now let's traverse the "old" list
        for local_customer in local_customers:

            # we differ if can't locate corresponding "new" host record
            host_customer = self.find_host_customer(host_customers, local_customer)
            if not host_customer:
                return True

        # guess we don't differ after all
        return False

    def find_local_customer(self, local_customers, host_customer):
        assert 'customerID' in self.supported_customer_fields

        if not host_customer['customerID']:
            return              # new customer

        for local_customer in local_customers:
            if str(local_customer['customerID']) == str(host_customer['customerID']):
                return local_customer

    def find_host_customer(self, host_customers, local_customer):
        assert 'customerID' in self.supported_customer_fields

        for host_customer in host_customers:
            if str(host_customer['customerID']) == str(local_customer['customerID']):
                return host_customer

    def create_object(self, key, data):
        # we can get away with using the same logic for both here
        return self.update_object(None, data)

    def update_object(self, member, data, local_data=None):
        """
        Push an update for the member, via the CORE API.
        """
        if self.dry_run:
            return data

        cardNo = data.pop('cardNo')
        # data = dict(data)
        # if data.get('startDate') == '__omit__':
        #     data.pop('startDate')
        # if data.get('endDate') == '__omit__':
        #     data.pop('endDate')
        member = self.api.set_member(cardNo, **data)
        return member


class DepartmentImporter(ToCoreAPI):
    """
    Department model importer for CORE-POS
    """
    model_name = 'Department'
    key = 'dept_no'
    supported_fields = [
        'dept_no',
        'dept_name',
        # TODO: should enable some of these fields?
        # 'dept_tax',
        # 'dept_fs',
        # 'dept_limit',
        # 'dept_minimum',
        # 'dept_discount',
        # 'dept_see_id',
        # 'modified',
        # 'modifiedby',
        # 'margin',
        # 'salesCode',
        # 'memberOnly',
    ]

    def get_local_objects(self, host_data=None):
        return self.api.get_departments()

    def get_single_local_object(self, key):
        assert len(self.key) == 1
        assert self.key[0] == 'dept_no'
        return self.api.get_department(key[0])

    def normalize_local_object(self, department):
        data = dict(department)
        return data

    def create_object(self, key, data):
        # we can get away with using the same logic for both here
        return self.update_object(None, data)

    def update_object(self, department, data, local_data=None):
        """
        Push an update for the department, via the CORE API.
        """
        if self.dry_run:
            return data

        dept_no = data.pop('dept_no')
        department = self.api.set_department(dept_no, **data)
        return department


class SubdepartmentImporter(ToCoreAPI):
    """
    Subdepartment model importer for CORE-POS
    """
    model_name = 'Subdepartment'
    key = 'subdept_no'
    supported_fields = [
        'subdept_no',
        'subdept_name',
        'dept_ID',
    ]

    def get_local_objects(self, host_data=None):
        return self.api.get_subdepartments()

    def get_single_local_object(self, key):
        assert len(self.key) == 1
        assert self.key[0] == 'subdept_no'
        return self.api.get_subdepartment(key[0])

    def normalize_local_object(self, subdepartment):
        data = dict(subdepartment)
        self.ensure_fields(data)
        return data

    def create_object(self, key, data):
        # we can get away with using the same logic for both here
        return self.update_object(None, data)

    def update_object(self, subdepartment, data, local_data=None):
        """
        Push an update for the subdepartment, via the CORE API.
        """
        if self.dry_run:
            return data

        subdept_no = data.pop('subdept_no')
        subdepartment = self.api.set_subdepartment(subdept_no, **data)
        return subdepartment


class VendorImporter(ToCoreAPI):
    """
    Vendor model importer for CORE-POS
    """
    model_name = 'Vendor'
    key = 'vendorID'
    supported_fields = [
        'vendorID',
        'vendorName',
        'vendorAbbreviation',
        'shippingMarkup',
        'discountRate',
        'phone',
        'fax',
        'email',
        'website',
        'address',
        'city',
        'state',
        'zip',
        'notes',
        'localOriginID',
        'inactive',
        'orderMinimum',
        'halfCases',
    ]

    def get_local_objects(self, host_data=None):
        return self.api.get_vendors()

    def get_single_local_object(self, key):
        assert len(self.key) == 1
        assert self.key[0] == 'vendorID'
        return self.api.get_vendor(key[0])

    def normalize_local_object(self, vendor):
        data = dict(vendor)

        # make sure all fields are present
        self.ensure_fields(data)

        # fix some "empty" values
        self.fix_empties(data, ['phone', 'fax', 'email'])

        # convert some values to native type
        data['discountRate'] = float(data['discountRate'])

        return data

    def create_object(self, key, data):
        # we can get away with using the same logic for both here
        return self.update_object(None, data)

    def update_object(self, vendor, data, local_data=None):
        """
        Push an update for the vendor, via the CORE API.
        """
        if self.dry_run:
            return data

        vendorID = data.pop('vendorID')
        vendor = self.api.set_vendor(vendorID, **data)
        return vendor


class ProductImporter(ToCoreAPI):
    """
    Product model importer for CORE-POS
    """
    model_name = 'Product'
    key = 'upc'
    supported_fields = [
        'upc',
        'brand',
        'description',
        'size',
        'department',
        'normal_price',
        'foodstamp',
        'scale',
        # 'tax',        # TODO!

        # TODO: maybe enable some of these fields?
        # 'formatted_name',
        # 'pricemethod',
        # 'groupprice',
        # 'quantity',
        # 'special_price',
        # 'specialpricemethod',
        # 'specialgroupprice',
        # 'specialquantity',
        # 'start_date',
        # 'end_date',
        # 'scaleprice',
        # 'mixmatchcode',
        # 'modified',
        # 'tareweight',
        # 'discount',
        # 'discounttype',
        # 'line_item_discountable',
        # 'unitofmeasure',
        # 'wicable',
        # 'qttyEnforced',
        # 'idEnforced',
        # 'cost',
        # 'inUse',
        # 'numflag',
        # 'subdept',
        # 'deposit',
        # 'local',
        # 'store_id',
        # 'default_vendor_id',
        # 'current_origin_id',
    ]

    def get_local_objects(self, host_data=None):
        return self.api.get_products()

    def get_single_local_object(self, key):
        assert len(self.key) == 1
        assert self.key[0] == 'upc'
        return self.api.get_product(key[0])

    def normalize_local_object(self, product):
        data = dict(product)

        # make sure all fields are present
        self.ensure_fields(data)

        # fix some "empty" values
        self.fix_empties(data, ['brand'])

        return data

    def create_object(self, key, data):
        # we can get away with using the same logic for both here
        return self.update_object(None, data)

    def update_object(self, product, data, local_data=None):
        """
        Push an update for the product, via the CORE API.
        """
        if self.dry_run:
            return data

        upc = data.pop('upc')
        product = self.api.set_product(upc, **data)
        return product



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE POS (API) -> Rattail data importing
"""

import datetime
import decimal
import logging
from collections import OrderedDict

from sqlalchemy import orm

from rattail import importing
from rattail.gpc import GPC
from rattail.time import localtime, make_utc
from rattail.core import get_uuid
from rattail.db.util import normalize_full_name
from rattail_corepos import importing as corepos_importing
from rattail_corepos.corepos.util import get_core_members
from rattail_corepos.corepos.api import make_corepos_api


log = logging.getLogger(__name__)


class FromCOREPOSToRattail(importing.ToRattailHandler):
    """
    Import handler for data coming from a CORE POS API.
    """
    host_key = 'corepos_api'
    generic_host_title = "CORE Office (API)"
    host_title = "CORE-POS (API)"
    safe_for_web_app = True

    def get_importers(self):
        importers = OrderedDict()
        importers['Customer'] = CustomerImporter
        importers['CustomerShopper'] = CustomerShopperImporter
        importers['MembershipType'] = MembershipTypeImporter
        importers['Member'] = MemberImporter
        importers['Store'] = StoreImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Vendor'] = VendorImporter
        importers['Product'] = ProductImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductMovement'] = ProductMovementImporter
        return importers

    def get_default_keys(self):
        keys = super(FromCOREPOSToRattail, self).get_default_keys()

        if 'ProductMovement' in keys:
            keys.remove('ProductMovement')
        return keys

    def get_core_members(self, api, progress=None):
        if not hasattr(self, 'cached_core_members'):
            self.cached_core_members = get_core_members(
                self.config, api, progress=progress)
        return self.cached_core_members


class FromCOREPOSAPI(importing.Importer):
    """
    Base class for all CORE POS API data importers.
    """

    def setup(self):
        super().setup()

        self.establish_api()

        self.ignore_new_members = self.should_ignore_new_members()

    def datasync_setup(self):
        super().datasync_setup()

        self.establish_api()

    def establish_api(self):
        self.api = make_corepos_api(self.config)

    def should_ignore_new_members(self):
        if hasattr(self, 'ignore_new_members'):
            return self.ignore_new_members

        return self.config.getbool('rattail_corepos',
                                   'importing_ignore_new_members',
                                   default=False)

    def get_core_members(self):
        members = self.handler.get_core_members(self.api, progress=self.progress)

        # maybe ignore NEW MEMBER accounts
        if self.should_ignore_new_members():
            members = [member for member in members
                       if not self.is_new_member(member)]

        return members

    def is_new_member(self, member):
        """
        Convenience method to check if the given member represents a
        "NEW MEMBER" record in CORE-POS, and hence it should be
        ignored for the import.
        """
        customers = member['customers']
        if customers:
            customer = customers[0]
            if (not customer['firstName']
                and customer['lastName'] == 'NEW MEMBER'):
                return True
        return False


class CustomerImporter(FromCOREPOSAPI, corepos_importing.model.CustomerImporter):
    """
    Importer for customer data from CORE POS API.
    """
    key = 'corepos_card_number'
    supported_fields = [
        'corepos_card_number',
        'corepos_account_id',
        'number',
        'name',
        'address_street',
        'address_street2',
        'address_city',
        'address_state',
        'address_zipcode',
    ]

    def get_host_objects(self):
        return self.get_core_members()

    def normalize_host_object(self, member):
        card_number = int(member['cardNo'])

        # figure out the "account holder" customer for the member.  note that
        # we only use this to determine the `Customer.name` in Rattail
        customers = member['customers']
        account_holders = [customer for customer in customers
                           if customer['accountHolder']]
        if account_holders:
            if len(account_holders) > 1:
                log.warning("member %s has %s account holders in CORE: %s",
                            member['cardNo'], len(account_holders), member)
            customer = account_holders[0]
        elif customers:
            if len(customers) > 1:
                log.warning("member %s has %s customers but no account holders: %s",
                            member['cardNo'], len(customers), member)
            customer = customers[0]
        else:
            raise NotImplementedError("TODO: how to handle member with no customers?")

        data = {
            'corepos_card_number': card_number,
            'corepos_account_id': int(member['customerAccountID']),
            'number': card_number,
            'name': normalize_full_name(customer['firstName'],
                                        customer['lastName']),

            'address_street': member['addressFirstLine'] or None,
            'address_street2': member['addressSecondLine'] or None,
            'address_city': member['city'] or None,
            'address_state': member['state'] or None,
            'address_zipcode': member['zip'] or None,
        }

        return data


class CustomerShopperImporter(FromCOREPOSAPI, corepos_importing.model.CustomerShopperImporter):
    """
    Importer for customer shopper data from CORE POS API.
    """
    key = ('customer_uuid', 'shopper_number')
    supported_fields = [
        'customer_uuid',
        'shopper_number',
        'corepos_customer_id',
        'first_name',
        'last_name',
        'display_name',
        'active',
        'phone_number',
        'phone_number_2',
        'email_address',
        'account_holder',
    ]

    def setup(self):
        super().setup()
        model = self.model

        self.maxlen_phone_number = self.get_maxlen_phone_number()

        self.customers_by_card_number = self.app.cache_model(
            self.session,
            model.Customer,
            key='corepos_card_number',
            query_options=[orm.joinedload(model.Customer._corepos)])

    def get_maxlen_phone_number(self):
        if hasattr(self, 'maxlen_phone_number'):
            return self.maxlen_phone_number

        model = self.model
        return self.app.maxlen(model.PhoneNumber.number)

    def get_host_objects(self):

        # first get all member data from CORE API
        members = self.get_core_members()
        normalized = []

        # then collect all the "shopper" records
        def normalize(member, i):
            normalized.extend(self.get_shoppers_for_member(member))

        self.progress_loop(normalize, members,
                           message="Collecting Person data from CORE")
        return normalized

    def get_customer_by_card_number(self, card_number):
        if hasattr(self, 'customers_by_card_number'):
            return self.customers_by_card_number.get(card_number)

        model = self.model
        try:
            return self.session.query(model.Customer)\
                               .join(model.CoreCustomer)\
                               .filter(model.CoreCustomer.corepos_card_number == card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_shoppers_for_member(self, member):
        """
        Return a list of shopper info dicts associated with the given
        member info dict (latter having come from the CORE API).
        """
        customers = member['customers']
        shoppers = []

        # make sure account holder is listed first
        account_holder = None
        secondary = False
        mixedup = False
        for customer in customers:
            if customer['accountHolder'] and not secondary:
                account_holder = customer
            elif not customer['accountHolder']:
                secondary = True
            elif customer['accountHolder'] and secondary:
                mixedup = True
        if mixedup:
            raise NotImplementedError("TODO: should re-sort the customers list for member {}".format(member['cardNo']))

        for i, customer in enumerate(customers, 1):
            shopper = dict(customer)
            shopper['card_number'] = member['cardNo']
            shopper['shopper_number'] = i
            shoppers.append(shopper)

        return shoppers

    def normalize_host_object(self, shopper):
        card_number = shopper['card_number']

        customer = self.get_customer_by_card_number(card_number)
        if not customer:
            log.warning("Rattail customer not found for CardNo %s: %s",
                        card_number, shopper)
            return

        data = {
            'customer_uuid': customer.uuid,
            'shopper_number': shopper['shopper_number'],
            'corepos_customer_id': int(shopper['customerID']),

            'first_name': shopper['firstName'],
            'last_name': shopper['lastName'],
            'display_name': normalize_full_name(shopper['firstName'],
                                                shopper['lastName']),

            # TODO: can a CORE shopper be *not* active?
            'active': True,

            'phone_number': shopper['phone'] or None,
            'phone_number_2': shopper['altPhone'] or None,

            'email_address': shopper['email'] or None,
            'account_holder': bool(shopper['accountHolder']),
        }

        # truncate phone number data if needed
        maxlen_phone_number = self.get_maxlen_phone_number()
        if data['phone_number'] and len(data['phone_number']) > maxlen_phone_number:
            log.warning("phone_number is too long (%s chars), "
                        "will truncate to %s chars: %s",
                        len(data['phone_number']),
                        maxlen_phone_number,
                        data['phone_number'])
            data['phone_number'] = data['phone_number'][:maxlen_phone_number]
        if data['phone_number_2'] and len(data['phone_number_2']) > maxlen_phone_number:
            log.warning("phone_number_2 is too long (%s chars), "
                        "will truncate to %s chars: %s",
                        len(data['phone_number_2']),
                        maxlen_phone_number,
                        data['phone_number_2'])
            data['phone_number_2'] = data['phone_number_2'][:maxlen_phone_number]

        # swap 1st and 2nd phone numbers if only latter has value
        self.prioritize_2(data, 'phone_number')

        return data


class StoreImporter(FromCOREPOSAPI, corepos_importing.model.StoreImporter):
    """
    Importer for store data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'id',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_stores()

    def normalize_host_object(self, store):
        return {
            'corepos_id': int(store['storeID']),
            'id': str(store['storeID']),
            'name': store['description'],
        }


class DepartmentImporter(FromCOREPOSAPI, corepos_importing.model.DepartmentImporter):
    """
    Importer for department data from CORE POS API.
    """
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_departments()

    def normalize_host_object(self, department):
        return {
            'corepos_number': int(department['dept_no']),
            'number': int(department['dept_no']),
            'name': department['dept_name'],
        }


class SubdepartmentImporter(FromCOREPOSAPI, corepos_importing.model.SubdepartmentImporter):
    """
    Importer for subdepartment data from CORE POS API.
    """
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
        'department_number',
    ]

    def get_host_objects(self):
        return self.api.get_subdepartments()

    def normalize_host_object(self, subdepartment):
        department_number = None
        if 'dept_ID' in subdepartment:
            department_number = int(subdepartment['dept_ID'])

        return {
            'corepos_number': int(subdepartment['subdept_no']),
            'number': int(subdepartment['subdept_no']),
            'name': subdepartment['subdept_name'],
            'department_number': department_number,
        }


class VendorImporter(FromCOREPOSAPI, corepos_importing.model.VendorImporter):
    """
    Importer for vendor data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'id',
        'name',
        'abbreviation',
        'special_discount',
        'phone_number',
        'fax_number',
        'email_address',
    ]

    def get_host_objects(self):
        return self.api.get_vendors()

    def normalize_host_object(self, vendor):
        return {
            'corepos_id': int(vendor['vendorID']),
            'id': str(vendor['vendorID']),
            'name': vendor['vendorName'],
            'abbreviation': vendor.get('vendorAbbreviation') or None,
            'special_discount': decimal.Decimal(vendor['discountRate']),
            'phone_number': vendor.get('phone') or None,
            'fax_number': vendor.get('fax') or None,
            'email_address': vendor.get('email') or None,
        }


class ProductImporter(FromCOREPOSAPI, corepos_importing.model.ProductImporter):
    """
    Importer for product data from CORE POS API.
    """
    key = 'uuid'
    supported_fields = [
        'uuid',
        'corepos_id',
        'item_id',
        'upc',
        'brand_name',
        'description',
        'unit_size',
        'unit_of_measure',
        'uom_abbreviation',
        'size',
        'weighed',
        'department_number',
        'subdepartment_number',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_type',
        'food_stampable',
        # 'tax1',
        # 'tax2',
        'case_size',
    ]

    def setup(self):
        super(ProductImporter, self).setup()
        model = self.model

        query = self.session.query(model.Product)\
                            .join(model.CoreProduct)\
                            .filter(model.CoreProduct.corepos_id != None)\
                            .options(orm.joinedload(model.Product._corepos))
        self.core_existing = self.app.cache_model(self.session,
                                                  model.Product,
                                                  key='corepos_id',
                                                  query=query)

        self.vendor_items_by_upc = {}

        def cache(item, i):
            self.vendor_items_by_upc.setdefault(item['upc'], []).append(item)

        self.progress_loop(cache, self.api.get_vendor_items(),
                           message="Caching CORE Vendor Items")

    def get_host_objects(self):
        return self.api.get_products()

    def identify_product(self, corepos_product):
        model = self.config.get_model()
        corepos_id = int(corepos_product['id'])

        if hasattr(self, 'core_existing'):
            product = self.core_existing.get(corepos_id)
            if product:
                return product

        else:
            try:
                return self.session.query(model.Product)\
                                   .join(model.CoreProduct)\
                                   .filter(model.CoreProduct.corepos_id == corepos_id)\
                                   .one()
            except orm.exc.NoResultFound:
                pass

        # at this point we'll search by `Product.item_id` instead
        return self.session.query(model.Product)\
                           .outerjoin(model.CoreProduct)\
                           .filter(model.CoreProduct.corepos_id == None)\
                           .filter(model.Product.item_id == corepos_product['upc'])\
                           .first()

    def identify_product_uuid(self, corepos_product):
        product = self.identify_product(corepos_product)
        if product:
            return product.uuid
        return get_uuid()

    def get_vendor_items(self, api_product):
        if hasattr(self, 'vendor_items_by_upc'):
            return self.vendor_items_by_upc.get(api_product['upc'])

        return self.api.get_vendor_items(upc=api_product['upc'])

    def normalize_host_object(self, product):
        try:
            upc = GPC(product['upc'], calc_check_digit='upc')
        except (TypeError, ValueError):
            log.debug("CORE POS product has invalid UPC: %s", product['upc'])
            if len(self.key) == 1 and self.key[0] == 'upc':
                return
            upc = None

        department_number = None
        if 'department' in product:
            department_number = int(product['department']) or None

        subdepartment_number = None
        if 'subdept' in product:
            subdepartment_number = int(product['subdept']) or None

        price = None
        if product.get('normal_price') is not None:
            price = decimal.Decimal(product['normal_price'])

        data = {
            'uuid': self.identify_product_uuid(product),
            'corepos_id': int(product['id']),
            'item_id': product['upc'],
            'upc': upc,
            'brand_name': product.get('brand') or None,
            'description': product.get('description') or '',

            'department_number': department_number,
            'subdepartment_number': subdepartment_number,

            'weighed': product.get('scale') == '1',
            'food_stampable': None,
            # 'tax1': product['tax'] == '1', # TODO: is this right?
            # 'tax2': product['tax'] == '2', # TODO: is this right?

            'regular_price_price': price,
            'regular_price_multiple': 1 if price is not None else None,
            'regular_price_type': self.enum.PRICE_TYPE_REGULAR if price is not None else None,
        }

        if 'foodstamp' in product:
            data['food_stampable'] = product['foodstamp'] == '1'

        if self.fields_active(self.size_fields):
            size_info = self.normalize_size_info(product)
            data.update({
                'size': size_info['size'],
                'unit_size': size_info['unit_size'],
                'unit_of_measure': size_info['uom_code'],
                'uom_abbreviation': size_info['uom_abbrev'],
            })

        if 'case_size' in self.fields:
            case_size = None
            items = self.get_vendor_items(product)
            if items:
                # here we sort by `modified DESC` to get the "deterministic
                # pseudo-default" vendorItems record

                def sortkey(item):
                    modified = item.get('modified')
                    if not modified:
                        return datetime.datetime(1900, 1, 1)
                    dt = datetime.datetime.strptime(modified, '%Y-%m-%d %H:%M:%S')
                    return dt

                items = sorted(items, key=sortkey, reverse=True)
                item = items[0]
                case_size = decimal.Decimal(item['units'])

            data['case_size'] = case_size

        return data


class ProductMovementImporter(FromCOREPOSAPI, corepos_importing.model.ProductImporter):
    """
    Importer for product movement data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'last_sold',
    ]
    allow_create = False
    allow_delete = False

    def get_host_objects(self):
        return self.api.get_products()

    def normalize_host_object(self, product):

        last_sold = None
        if 'last_sold' in product:
            last_sold = datetime.datetime.strptime(product['last_sold'], '%Y-%m-%d %H:%M:%S')
            last_sold = localtime(self.config, last_sold)
            last_sold = make_utc(last_sold)

        return {
            'corepos_id': int(product['id']),
            'last_sold': last_sold,
        }


class ProductCostImporter(FromCOREPOSAPI, corepos_importing.model.ProductCostImporter):
    """
    Importer for product cost data from CORE POS API.
    """
    key = ('vendor_uuid', 'code')
    supported_fields = [
        'corepos_id',
        'product_uuid',
        'vendor_uuid',
        'code',
        'case_size',
        'case_cost',
        'unit_cost',
        'preferred',
    ]

    def setup(self):
        super(ProductCostImporter, self).setup()
        model = self.config.get_model()

        query = self.session.query(model.Vendor)\
                            .join(model.CoreVendor)\
                            .filter(model.CoreVendor.corepos_id != None)
        self.vendors = self.app.cache_model(self.session,
                                            model.Vendor,
                                            query=query,
                                            key='corepos_id')

        self.corepos_products = {}

        def cache(product, i):
            self.corepos_products[product['upc']] = product

        self.progress_loop(cache, self.api.get_products(),
                           message="Caching Products from CORE-POS API")

        self.products_by_item_id = self.app.cache_model(self.session,
                                                        model.Product,
                                                        key='item_id')

    def get_host_objects(self):
        return self.api.get_vendor_items()

    def get_vendor(self, item):
        corepos_id = int(item['vendorID'])

        if hasattr(self, 'vendors'):
            return self.vendors.get(corepos_id)

        model = self.config.get_model()
        try:
            return self.session.query(model.Vendor)\
                               .join(model.CoreVendor)\
                               .filter(model.CoreVendor.corepos_id == corepos_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_corepos_product(self, item):
        if hasattr(self, 'corepos_products'):
            return self.corepos_products.get(item['upc'])

        return self.api.get_product(item['upc'])

    def get_product(self, item):
        item_id = item['upc']

        if hasattr(self, 'products_by_item_id'):
            return self.products_by_item_id.get(item_id)

        model = self.model
        try:
            return self.session.query(model.Product)\
                               .filter(model.Product.item_id == item_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, item):
        vendor = self.get_vendor(item)
        if not vendor:
            log.warning("CORE POS vendor not found for item: %s", item)
            return

        product = self.get_product(item)
        if not product:
            # just debug logging since this is a common scenario; the
            # CORE table is for items "available from vendor" but not
            # necssarily items carried by store
            log.debug("product not found for CORE vendor item: %s", item)
            return

        core_product = self.get_corepos_product(item)
        # if not product:
        #     log.warning("CORE POS product not found for item: %s", item)
        #     return

        preferred = False
        if core_product and core_product['default_vendor_id'] == item['vendorID']:
            preferred = True

        case_size = decimal.Decimal(item['units'])
        unit_cost = item.get('cost')
        if unit_cost is not None:
            unit_cost = decimal.Decimal(unit_cost)
        case_cost = None
        if unit_cost is not None:
            case_cost = unit_cost * case_size

        return {
            'corepos_id': int(item['vendorItemID']),
            'product_uuid': product.uuid,
            'vendor_uuid': vendor.uuid,
            'code': (item['sku'] or '').strip() or None,
            'case_size': case_size,
            'case_cost': case_cost,
            'unit_cost': unit_cost,
            'preferred': preferred,
        }


class MembershipTypeImporter(FromCOREPOSAPI, importing.model.MembershipTypeImporter):
    """
    Imports membership type data from CORE-POS API
    """
    key = 'number'
    supported_fields = [
        'number',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_member_types()

    def normalize_host_object(self, memtype):
        return {
            'number': int(memtype['memtype']),
            'name': memtype['memDesc'],
        }


class MemberImporter(FromCOREPOSAPI, corepos_importing.model.MemberImporter):
    """
    Importer for member data from CORE POS API.
    """
    key = 'number'
    supported_fields = [
        'number',
        'corepos_account_id',
        'customer_uuid',
        'person_first_name',
        'person_last_name',
        'membership_type_number',
        'joined',
        'withdrew',
        'active',
    ]

    # TODO: should make this configurable
    member_status_codes = [
        'PC',
        'TERM',
    ]

    # TODO: should make this configurable
    non_member_status_codes = [
        'REG',
        'INACT',
    ]

    def setup(self):
        super().setup()
        model = self.model

        self.customers_by_number = self.app.cache_model(
            self.session,
            model.Customer,
            key='number')

    def get_host_objects(self):
        return self.get_core_members()

    def get_customer_by_number(self, number):
        if hasattr(self, 'customers_by_number'):
            return self.customers_by_number.get(number)

        model = self.model
        try:
            return self.session.query(model.Customer)\
                               .filter(model.Customer.number == number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_corepos_customer(self, corepos_member):
        for customer in corepos_member['customers']:
            if customer['accountHolder']:
                return customer

    def normalize_host_object(self, member):
        card_number = member['cardNo']
        customer = self.get_customer_by_number(card_number)

        # TODO: at first i was *skipping* non-member status records,
        # but since CORE sort of assumes all customers are members,
        # probably not worth making the distinction here..?  it is
        # important to import the full member info from CORE, so that
        # we have it to sync back.  therefore can't afford to "skip"
        # any member records here
        if (member['memberStatus'] not in self.member_status_codes
            and member['memberStatus'] not in self.non_member_status_codes):
            log.warning("unexpected status '%s' for member %s: %s",
                        member['memberStatus'], card_number, member)

        joined = None
        if member['startDate'] and member['startDate'] != '0000-00-00 00:00:00':
            joined = datetime.datetime.strptime(member['startDate'],
                                                '%Y-%m-%d %H:%M:%S')
            joined = joined.date()

        withdrew = None
        if (member['endDate']
            and member['endDate'] != '0000-00-00 00:00:00'
            and member['endDate'] != '1900-01-01 00:00:00'):
            withdrew = datetime.datetime.strptime(member['endDate'],
                                                  '%Y-%m-%d %H:%M:%S')
            withdrew = withdrew.date()

        typeno = int(member['customerTypeID']) or None
        if typeno:
            memtype = self.get_membership_type_by_number(typeno)
            if not memtype:
                logger = log.warning if self.get_warn_for_unknown_membership_type() else log.debug
                logger("unknown customerTypeID (membership_type_number) %s for: %s",
                       member['customerTypeID'], member)
                typeno = None

        data = {
            'number': card_number,
            'corepos_account_id': int(member['customerAccountID']),
            'customer_uuid': customer.uuid if customer else None,
            'person_first_name': None,
            'person_last_name': None,
            'membership_type_number': typeno,
            'joined': joined,
            'withdrew': withdrew,
            'active': not bool(withdrew),
        }

        corepos_customer = self.get_corepos_customer(member)
        if corepos_customer:
            data['person_first_name'] = corepos_customer['firstName']
            data['person_last_name'] = corepos_customer['lastName']

        return data



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Email profiles for Rattail / CORE-POS integration
"""

from rattail.emails import ImporterEmail, ProblemReportEmail


class core_office_export_lane_op_updates(ImporterEmail):
    """
    Sent when a CORE Office -> CORE Lane export involves data changes.
    """
    handler_spec = 'rattail_corepos.corepos.lane.importing.op.office:FromCoreOfficeToCoreLane'
    abstract = False


class core_office_import_csv_updates(ImporterEmail):
    """
    Sent when CSV -> CORE import involves data changes.
    """
    handler_spec = 'rattail_corepos.corepos.office.importing.db.csv:FromCSVToCore'
    abstract = False


class corepos_problems_invalid_person_numbers(ProblemReportEmail):
    """
    Looks for `custdata` records with invalid person number sequence.
    """
    default_subject = "Invalid person numbers"
    abstract = False

    def sample_data(self, request):
        from corepos.db.office_op import model as corepos

        customer = corepos.CustomerClassic(card_number=42,
                                    first_name="Fred",
                                    last_name="Flintstone",
                                    person_number=2)
        return {
            'problems': [(customer, 1)]
        }


class corepos_problems_phone_numbers_too_long(ProblemReportEmail):
    """
    Looks for `meminfo` records with phone number which is too long to
    properly fit in the `Customers` table.
    """
    default_subject = "Phone numbers too long"
    abstract = False

    def sample_data(self, request):
        from corepos.db.office_op import model as corepos

        member = corepos.MemberInfo(card_number=42,
                                    phone='(800) 555-1234 ABCDEFGHIJKLMNOP')
        return {
            'problems': [member]
        }


class rattail_export_corepos_updates(ImporterEmail):
    """
    Sent when a Rattail -> CORE-POS API export involves data changes.
    """
    handler_spec = 'rattail_corepos.corepos.office.importing.rattail:FromRattailToCore'
    abstract = False


class rattail_import_corepos_api_updates(ImporterEmail):
    """
    Sent when a CORE-POS API -> Rattail import involves data changes.
    """
    handler_spec = 'rattail_corepos.importing.corepos.api:FromCOREPOSToRattail'
    abstract = False


class rattail_import_corepos_db_updates(ImporterEmail):
    """
    Sent when a CORE-POS DB -> Rattail import involves data changes.
    """
    handler_spec = 'rattail_corepos.importing.corepos.db:FromCOREPOSToRattail'
    abstract = False



# -*- coding: utf-8 -*-
"""add Person extension

Revision ID: c61f78243ff3
Revises: b43e93d32275
Create Date: 2020-03-16 16:50:55.266325

"""

# revision identifiers, used by Alembic.
revision = 'c61f78243ff3'
down_revision = 'b43e93d32275'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_person
    op.create_table('corepos_person',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_customer_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['person.uuid'], name='corepos_person_fk_person'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_person_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_customer_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_person_version_end_transaction_id'), 'corepos_person_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_person_version_operation_type'), 'corepos_person_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_person_version_transaction_id'), 'corepos_person_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_person
    op.drop_index(op.f('ix_corepos_person_version_transaction_id'), table_name='corepos_person_version')
    op.drop_index(op.f('ix_corepos_person_version_operation_type'), table_name='corepos_person_version')
    op.drop_index(op.f('ix_corepos_person_version_end_transaction_id'), table_name='corepos_person_version')
    op.drop_table('corepos_person_version')
    op.drop_table('corepos_person')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office commands
"""

import logging
import sys

import requests
from requests.auth import HTTPDigestAuth

from rattail import commands
from rattail_corepos import __version__
from rattail.util import load_object
from rattail_corepos.corepos.office.util import get_fannie_config_value
from rattail_corepos.corepos.api import make_corepos_api
from rattail_corepos.corepos.util import get_core_members
from rattail_corepos.config import core_office_url


log = logging.getLogger(__name__)


def main(*args):
    """
    Entry point for 'core-office' commands
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)


class Command(commands.Command):
    """
    Primary command for CORE Office
    """
    name = 'core-office'
    version = __version__
    description = "core-office -- command line interface for CORE Office"
    long_description = ""


class Anonymize(commands.Subcommand):
    """
    Update customer etc. data to make it anonymous
    """
    name = 'anonymize'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--force', '-f', action='store_true',
                            help="Do not prompt for confirmation.")

    def run(self, args):
        if not args.force:
            self.rprint("\n[bold yellow]**WARNING** this will modify all customer (and similar) records![/bold yellow]")
            value = input("\nreally want to do this? [yN] ")
            if not value or not self.config.parse_bool(value):
                self.stderr.write("user canceled\n")
                sys.exit(1)

        try:
            import names
        except ImportError:
            self.stderr.write("must install the `names` package first!\n\n"
                              "\tpip install names\n")
            sys.exit(2)

        from rattail_corepos.corepos.api import make_corepos_api
        from rattail_corepos.corepos.util import get_core_members

        api = make_corepos_api(self.config)
        members = get_core_members(self.config, api, progress=self.progress)

        def anonymize(member, i):
            data = dict(member)
            cardno = data.pop('cardNo')

            for customer in data['customers']:
                customer['firstName'] = names.get_first_name()
                customer['lastName'] = names.get_last_name()

            api.set_member(cardno, **data)

        self.progress_loop(anonymize, members,
                           message="Anonymizing all member data")


class CoreDBImportSubcommand(commands.ImportSubcommand):
    """
    Base class for commands which import straight to CORE DB
    """

    def add_parser_args(self, parser):
        super().add_parser_args(parser)

        parser.add_argument('--corepos-dbtype', metavar='TYPE', default='office_op',
                            choices=['office_op', 'office_trans'],
                            help="Config *type* for CORE-POS database engine to which data "
                            "should be written.  Default type is 'office_op' - this determines "
                            "which config section is used with regard to the --corepos-dbkey arg.")

        parser.add_argument('--corepos-dbkey', metavar='KEY', default='default',
                            help="Config key for CORE-POS database engine to which data should "
                            "be written.  This key must be defined in the config section as "
                            "determiend by the --corpos-dbtype arg.")

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['corepos_dbtype'] = kwargs['args'].corepos_dbtype
            kwargs['corepos_dbkey'] = kwargs['args'].corepos_dbkey
        return kwargs


class ExportLaneOp(commands.ImportSubcommand):
    """
    Export "op" data from CORE Office to CORE Lane
    """
    name = 'export-lane-op'
    description = __doc__.strip()
    handler_key = 'to_corepos_db_lane_op.from_corepos_db_office_op.export'
    default_dbkey = 'default'

    def add_parser_args(self, parser):
        super(ExportLaneOp, self).add_parser_args(parser)
        parser.add_argument('--dbkey', metavar='KEY', default=self.default_dbkey,
                            help="Config key for database engine to be used as the "
                            "\"target\" CORE Lane DB, i.e. where data will be "
                            " exported.  This key must be defined in the "
                            " [rattail_corepos.db.lane_op] section of your "
                            "config file.")

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs


class GetConfigValue(commands.Subcommand):
    """
    Get a value from CORE Office `fannie/config.php`
    """
    name = 'get-config-value'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('name', metavar='NAME',
                            help="Name of the config value to get.  "
                            "Prefix of `FANNIE_` is not required.")

    def run(self, args):
        value = get_fannie_config_value(self.config, args.name)
        self.stdout.write(f"{value}\n")


class ExportCSV(commands.ExportFileSubcommand):
    """
    Export data from CORE to CSV file(s)
    """
    name = 'export-csv'
    description = __doc__.strip()
    handler_key = 'to_csv.from_corepos_db_office_op.export'


class ImportCSV(commands.ImportFileSubcommand):
    """
    Import data from CSV to CORE Office "op" DB
    """
    name = 'import-csv'
    description = __doc__.strip()
    handler_key = 'to_corepos_db_office_op.from_csv.import'

    def add_parser_args(self, parser):
        super().add_parser_args(parser)

        parser.add_argument('--dbkey', metavar='KEY', default='default',
                            help="Config key for database engine to be used as the \"target\" "
                            "CORE DB, i.e. where data will be imported *to*.  This key must be "
                            "defined in the [corepos.db.office_op] section of your config file.")

    def get_handler_kwargs(self, **kwargs):
        kwargs = super().get_handler_kwargs(**kwargs)

        if 'args' in kwargs:
            args = kwargs['args']
            kwargs['dbkey'] = args.dbkey

        return kwargs


class ImportSelf(commands.ImportSubcommand):
    """
    Import data from CORE Office ("op" DB) to "self"
    """
    name = 'import-self'
    description = __doc__.strip()
    handler_key = 'to_self.from_corepos_db_office_op.import'


class PatchCustomerGaps(commands.Subcommand):
    """
    POST to the CORE API as needed, to patch gaps for customerID
    """
    name = 'patch-customer-gaps'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--dry-run', action='store_true',
                            help="Do not POST anything, but log members needing it.")

    def run(self, args):
        from corepos.db.office_op import model as corepos

        corepos_api = make_corepos_api(self.config)
        members = get_core_members(self.config, corepos_api, progress=self.progress)
        tally = self.app.make_object(updated=0)

        self.maxlen_phone = self.app.maxlen(corepos.Customer.phone)
        # nb. just in case the smallest one changes in future..
        other = self.app.maxlen(corepos.MemberInfo.phone)
        if other < self.maxlen_phone:
            self.maxlen_phone = other

        def inspect(member, i):
            for customer in member['customers']:
                customer_id = int(customer['customerID'])
                if not customer_id:
                    data = dict(member)
                    self.trim_phones(data)
                    cardno = data.pop('cardNo')
                    log.debug("%s call set_member() for card no %s: %s",
                              'should' if args.dry_run else 'will',
                              cardno, data)
                    if not args.dry_run:
                        corepos_api.set_member(cardno, **data)
                    tally.updated += 1
                    return

        action = "Finding"
        if not args.dry_run:
            action += " and fixing"
        self.progress_loop(inspect, members,
                           message=f"{action} customerID gaps")

        self.stdout.write("\n")
        if args.dry_run:
            self.stdout.write("would have ")
        self.stdout.write(f"updated {tally.updated} members\n")

    def trim_phones(self, data):
        # the `meminfo` table allows 30 chars for phone, but
        # `Customers` table only allows 20 chars.  so we must trim to
        # 20 chars or else the CORE API will silently fail to update
        # tables correctly when we POST to it
        for customer in data['customers']:
            for field in ['phone', 'altPhone']:
                value = customer[field]
                if len(value) > self.maxlen_phone:
                    log.warning("phone value for cardno %s is too long (%s chars) "
                                "and will be trimmed to %s chars: %s",
                                data['cardNo'],
                                len(value),
                                self.maxlen_phone,
                                value)
                    customer[field] = value[:self.maxlen_phone]


class PingInstall(commands.Subcommand):
    """
    Ping the /install URL in CORE Office (for DB setup)
    """
    name = 'ping-install'
    description = __doc__.strip()

    def run(self, args):
        url = core_office_url(self.config, require=True)
        url = f'{url}/install/'

        # TODO: hacky re-using credentials from API config..
        username = self.config.get('corepos.api', 'htdigest.username')
        password = self.config.get('corepos.api', 'htdigest.password')

        session = requests.Session()
        if username and password:
            session.auth = HTTPDigestAuth(username, password)

        response = session.get(url)
        response.raise_for_status()
